{Another version of the Game of Life.  Based loosely on Michael Abrash's}
{version in his book Zen of Code Optimization.  The only optimization done}
{on this version is that of the calc_neighbors() routine. Look in life.asm}
{for more information}

{If you don't have ObjGfx or the Keys unit, you can download them from:
 http://www.monstersoft.com/objgfx/
 or directly from:
 http://www.monstersoft.com/download/objgfx.zip  }

{Written by TCA of NewOrder}
program Life;
uses
 ObjGfx0,ObjGfx8,Keys;

const ON_COLOUR=1;  {This MUST be 1!!!!}
const OFF_COLOUR=0; {This MUST be 0!!!!}

type TLIfeObj=object(TGfx8)
       width,height       : word;
       constructor init(H,V:Word);
       procedure next_generation(var next_map:TLifeObj);
       function calc_neighbors(x,y:word):byte;
       procedure copy_cells(var sourcemap:TLifeObj);
     end;

type PLifeObj=^TLifeObj;

{$L life} {life.asm}

function TLifeObj.calc_neighbors(x,y:word):byte; external;

constructor TLifeObj.init(H,V:word);
begin
 inherited init;
 width:=H;
 height:=v;

 {make the buffer 2 pixels wider than the actual area so that no clipping}
 {is required.  This version does not wrap, but could be made to do so.}
 if not getmem(width+2,height+2) then
  begin
   writeln('Error allocating memory');
   halt
  end;
end;

procedure TLifeObj.copy_cells(var sourcemap:TLifeObj);
begin
 {copy the source map}
 flip(sourcemap);

 {clear all cells around the edge of the cellmap}
 rectangle(0,0,getmaxx,getmaxy,OFF_COLOUR)
end;

procedure TLifeObj.next_generation(var next_map:TLifeOBj);
var x,y,neighbor_count:word;
begin
 for y:=1 to height-1 do
  for x:=1 to width-1 do
   begin
    {calculate the # of neighbors to this cell:}
    neighbor_count:=calc_neighbors(x,y);
    if getpixel(x,y)=ON_COLOUR then {is this cell on?}
     begin
      if (neighbor_count<>2) and (neighbor_count<>3) then
       next_map.putpixel(x,y,OFF_COLOUR)
     end
    else {this cell wasn't on}
     if neighbor_count=3 then next_map.putpixel(x,y,ON_COLOUR)
   end
end;

const cellmap_width:word=320;
const cellmap_height:word=200;

var current_map,next_map:PLifeObj;
var Screen:TScreen8;
var init_length,x,y:word;
    generation:longint;

begin
 randomize;

 Screen.init;

 if not screen.setmode(_320x200x8bpp) then {do some basic error checking}
  begin
   writeln('Error setting video mode');
   halt
  end;

 {uncomment out the next two lines make the cell_map the same size}
 {as the screen
 cellmap_width:=screen.getmaxx+1;
 cellmap_height:=screen.getmaxy+1; {}

 {init the two life cell maps}
 new(current_map,init(cellmap_width,cellmap_height));
 new(next_map,init(cellmap_width,cellmap_height));

 {randomly place "cells"}
 init_length:=(longint(cellmap_width)*cellmap_height) div 2;
 repeat
  x:=random(cellmap_width);
  y:=random(cellmap_height);
  next_map^.putpixel(x,y,ON_COLOUR);
  dec(init_length);
 until init_length=0;

 {center the cell field.  This only works if next_map^.width<=screen.getmaxx}
 {and next_map^.height<=screen.getmaxY}
 x:=(Screen.GetMaxX-next_map^.width+1) div 2;
 y:=(Screen.GetMaxY-next_map^.height+1) div 2;

 {main loop}
 repeat
   inc(generation);
   current_map^.next_generation(next_map^);
   current_map^.copy_cells(next_map^);
   screen.copybuf(x,y,next_map^,1,1,next_map^.getmaxx-1,next_map^.getmaxy-1);
 until keypressed;

 {de-allocate the TLifeObjects and call their destructors}
 dispose(current_map,done);
 dispose(next_map,done);

 {turn off the screen}
 Screen.Done;

 writeln('Generations: ',generation);
 while keypressed do readkey; {just in case of an over-zealous user}
end.